#!/usr/bin/env python
'''
component_tf
Created by Seria at 05/02/2019 1:41 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from functools import partial
from mxnet.gluon import nn, loss, Trainer, HybridBlock
from mxnet import nd, init, autograd, lr_scheduler
from math import ceil
import os
from ..toolkit.utility import recordConfig
from .component import Pod


class MXsigmoid(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsigmoid, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.sigmoid(input)

class MXtanh(HybridBlock):
    def __init__(self, **kwargs):
        super(MXtanh, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.tanh(input)

class MXsoftmax(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsoftmax, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.softmax(input)

class MXrelu(HybridBlock):
    def __init__(self, **kwargs):
        super(MXrelu, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.relu(input)

class Duplicate(HybridBlock):
    def __init__(self, **kwargs):
        super(Duplicate, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return input

class MXwd(HybridBlock):
    def __init__(self, **kwargs):
        super(MXwd, self).__init__(**kwargs)

    def hybrid_forward(self, F, penalty):
        return F.zeros((1))

class MXmse(HybridBlock):
    def __init__(self, **kwargs):
        super(MXmse, self).__init__(**kwargs)

    def hybrid_forward(self, F, input, label):
        return (input-label)**2

class MXmae(HybridBlock):
    def __init__(self, **kwargs):
        super(MXmae, self).__init__(**kwargs)

    def hybrid_forward(self, F, input, label):
        return F.abs(input-label)

class MXsigmxe(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsigmxe, self).__init__(**kwargs)
        self.cost = loss.SigmoidBCELoss()

    def hybrid_forward(self, F, input, label):
        return F.mean(self.cost(input, label))

class MXsftmxe(HybridBlock):
    def __init__(self, is_one_hot, **kwargs):
        super(MXsftmxe, self).__init__(**kwargs)
        self.cost = loss.SoftmaxCELoss(sparse_label=not is_one_hot)

    def hybrid_forward(self, F, input, label):
        return F.mean(self.cost(input, label))

class MX(HybridBlock):
    def __init__(self, **kwargs):
        super(MX, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return nd.sigmoid(input)

class AccuracyCLS(HybridBlock):
    def __init__(self, multi_class, is_one_hot, **kwargs):
        super(AccuracyCLS, self).__init__(**kwargs)
        if multi_class:
            assert not is_one_hot
        self.mulcls = multi_class
        self.onehot = is_one_hot

    def hybrid_forward(self, F, input, label):
        if self.mulcls:
            pred = F.round(input)
            correct = F.mean(pred == label, axis=-1)
            return F.mean(correct == 1)
        else:
            if self.onehot:
                label = F.argmax(label, axis=-1)
            pred = F.argmax(input, axis=-1)
            return F.mean(pred == label)

class ComponentMX(object):
    def __init__(self, channel_major=True, time_major=False):
        self.channel_major = channel_major
        self.time_major = time_major
        self.warehouse = ['CONV_1D', 'CONV', 'CONV_3D', 'CONV_TRANS', 'CONV_SEP',
                          'SIGMOID', 'TANH', 'SOFTMAX', 'RELU', 'RELU_LEAKY', 'RELU_EXP',
                          'MAX_POOL', 'AVG_POOL',
                          'DROPOUT', 'BATCH_NORM', 'EMBEDDING',
                          'RESHAPE', 'FLAT', 'SLICE',
                          'CLIP', 'DENSE',
                          'RESIZE',
                          'DUPLICATE', 'CONVERT',
                          'WEIGHT_DECAY', 'SIGM_XENTROPY', 'SFTM_XENTROPY', 'MSE', 'MAE',
                          'MOMENTUM', 'NESTEROV', 'ADAM']
        # Convolution
        self.CONV_1D = 0
        self.CONV = self.conv_2d
        self.CONV_3D = 2
        self.CONV_TRANS = 3
        self.CONV_SEP = 4
        # Activation
        self.SIGMOID = self.sigmoid
        self.TANH = self.tanh
        self.SOFTMAX = self.softmax
        self.RELU = self.relu
        self.RELU_LEAKY = self.relu_leaky
        # self.RELU_EXP = self.relu_exp
        # Pooling
        self.MAX_POOL = self.max_pool_2d
        self.AVG_POOL = self.avg_pool_2d
        # Distributing
        self.DROPOUT = self.dropout
        self.BATCH_NORM = self.batch_norm
        # self.EMBEDDING = self.embedding
        # Reshape
        # self.RESHAPE = self.reshape
        self.FLAT = self.flat
        # self.SLICE = self.slice
        # self.PAD = self.pad
        # Arithmetic
        # self.CLIP = self.clip
        self.DENSE = self.dense
        # self.ARGMAX = self.argmax
        # Image Manipulation
        # self.RESIZE = self.resize
        # Copy or Rename
        self.DUPLICATE = self.duplicate
        # self.CONVERT = self.convert
        # Loss
        self.WEIGHT_DECAY = self.weight_decay
        self.SIGM_XENTROPY = self.sigm_xentropy
        self.SFTM_XENTROPY = self.sftm_xentropy
        self.MSE = self.mse
        self.MAE = self.mae
        # Optimizer
        self.MOMENTUM = self.momentum
        self.NESTEROV = self.nesterov
        self.ADAM = self.adam
        # Metric
        self.ACC_CLS = self.accuracy_cls



    def addComp(self, name, comp, is_complete=False):
        name = name.upper()
        if name in self.warehouse:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)
        else:
            self.warehouse.append(name)
        if is_complete:
            def customizedComp(name):
                return Pod(partial(comp.component, name=name), comp.symbol, name, comp.message)
        else:
            assert not isinstance(comp.component, list)
            def customizedComp(**kwargs):
                message = []
                # check input so as to update message
                for s, sym in enumerate(comp.symbol):
                    if sym in kwargs:
                        message.append(kwargs[sym])
                    else:
                        message.append(comp.message[s])
                return Pod(partial(comp.component, **kwargs), comp.symbol, kwargs['name'], message)
        exec('self.%s = customizedComp' % name)

    # ------------------------------------ Convolution ------------------------------------ #

    def _getInitializer(self, initializer, param, regularizer):
        init_err = Exception('NEBULAE ERROR ⨷ %s initializer is not defined or supported.' % initializer)
        if initializer == 'xavier':
            if param is None:
                param = 'gaussian'
            else:
                if param=='normal':
                    param = 'gaussian'
            var = init.Xavier(param)
        elif initializer == 'uniform':
            if param is None:
                param = 0.5
            else:
                if param[0] != 0:
                    raise Exception('NEBULAE ERROR ⨷ mean of uniform distributed initializer must be 0 in MXNET core.')
                param = param[1]
            var = init.Uniform(param)
        elif initializer == 'normal':
            if param is None:
                param = 0.1
            else:
                if param[0]+param[1] != 0:
                    raise Exception('NEBULAE ERROR ⨷ min and max value of normal distributed initializer must be',
                                    'opposite numbers in MXNET core.')
                param = param[1]
            var = init.Normal(param)
        elif initializer == 'zero':
            var = init.Zero()
        elif initializer == 'one':
            var = init.One()
        else:
            raise init_err

        if regularizer == 'l2':
            pass
        else:
            raise Exception('NEBULAE ERROR ⨷ %s regularizer is not defined or supported.' % regularizer)

        return var

    def conv_2d(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._conv_2d, **kwargs), ['input'], kwargs['name'], [message])
    def _conv_2d(self, name, input, kernel, out_chs, w_init='xavier', w_param=None,
                 b_init=None, b_param=None, w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), keep_size=True):
        if self.channel_major:
            format = 'NCHW'
            h, w = input[0][2:]
        else:
            format = 'NHWC'
            h, w = input[0][1:3]
        if keep_size:
            y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] + (dilation[0] - 1) * (kernel[0] - 1) - h) / 2)
            x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] + (dilation[1] - 1) * (kernel[1] - 1) - w) / 2)
            padding = (y, x)
        else:
            padding = (0, 0)
        if not w_init is None:
            w_init = self._getInitializer(w_init, w_param, w_reg)
        if not b_init is None:
            use_bias = True
            b_init = self._getInitializer(b_init, b_param, b_reg)
        else:
            use_bias = False
        return nn.Conv2D(channels=out_chs, kernel_size=kernel, strides=stride, padding=padding, dilation=dilation,
                         layout=format, use_bias=use_bias, weight_initializer=w_init, bias_initializer=b_init)

    # ------------------------------------ Activation ------------------------------------ #

    def sigmoid(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._sigmoid, **kwargs), ['input'], kwargs['name'], [message])
    def _sigmoid(self, name, input):
        return MXsigmoid()

    def tanh(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._tanh, **kwargs), ['input'], kwargs['name'], [message])
    def _tanh(self, name, input):
        return MXtanh()

    def softmax(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._tanh, **kwargs), ['input'], kwargs['name'], [message])
    def _softmax(self, name, input):
        return MXsoftmax()

    def relu(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._relu, **kwargs), ['input'], kwargs['name'], [message])
    def _relu(self, name, input):
        return MXrelu()

    def relu_leaky(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._relu_leaky, **kwargs), ['input'], kwargs['name'], [message])
    def _relu_leaky(self, name, input, alpha=0.2):
        return nn.LeakyReLU(alpha)

    # ------------------------------------ Distributing ------------------------------------ #

    def dropout(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._dropout, **kwargs), ['input'], kwargs['name'], [message])
    def _dropout(self, name, input, p_drop):
        return nn.Dropout(p_drop)

    def batch_norm(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('is_train', '_IS_TRAIN'))
        return Pod(partial(self._batch_norm, **kwargs), ['input', 'is_train'], kwargs['name'], message)
    def _batch_norm(self, name, input, is_train, beta=False, gamma=False):
        if self.channel_major:
            axis = 1
        else:
            axis = -1
        return nn.BatchNorm(axis=axis, momentum=0.9, epsilon=1e-05, center=beta, scale=gamma)

    # def embedding(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._embedding, **kwargs), ['input'], kwargs['name'], [message])
    # def _embedding(self, name, input, vocabulary, vec_dims, w_init='xavier', w_param=None):
    #     embd_vec = self._createVar(name+'_vec', [vocabulary, vec_dims], w_init, w_param)
    #     return tf.nn.embedding_lookup(embd_vec, input, name=name)

    # ------------------------------------ Pooling ------------------------------------ #

    def max_pool_2d(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._max_pool_2d, **kwargs), ['input'], kwargs['name'], [message])
    def _max_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), keep_size=True, if_global=False):
        if self.channel_major:
            format = 'NCHW'
            h, w = input[0][2:]
        else:
            format = 'NHWC'
            h, w = input[0][1:3]
        if if_global:
            kernel = (h, w)
            padding = (0, 0)
        else:
            if keep_size:
                y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] - h) / 2)
                x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] - w) / 2)
                padding = (y, x)
            else:
                padding = (0, 0)
        return nn.MaxPool2D(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=format)

    def avg_pool_2d(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._avg_pool_2d, **kwargs), ['input'], kwargs['name'], [message])
    def _avg_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), keep_size=True, if_global=False):
        if self.channel_major:
            format = 'NCHW'
            h, w = input[0][2:]
        else:
            format = 'NHWC'
            h, w = input[0][1:3]
        if if_global:
            kernel = (h, w)
            padding = (0, 0)
        else:
            if keep_size:
                y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] - h) / 2)
                x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] - w) / 2)
                padding = (y, x)
            else:
                padding = (0, 0)
        return nn.AvgPool2D(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=format)

    # ------------------------------------ Reshape ------------------------------------ #

    # def reshape(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._reshape, **kwargs), ['input'], kwargs['name'], [message])
    # def _reshape(self, name, input, shape):
    #     return tf.reshape(input, shape, name)
    #
    def flat(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._flat, **kwargs), ['input'], kwargs['name'], [message])
    def _flat(self, name, input):
        return nn.Flatten()
    #
    # def slice(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._slice, **kwargs), ['input'], kwargs['name'], [message])
    # def _slice(self, name, input, indices):
    #     for d in range(len(indices)):
    #         output = input[indices[d][0]:indices[d][1]]
    #     return tf.identity(output, name)
    #
    # def pad(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._pad, **kwargs), ['input'], kwargs['name'], [message])
    # def _pad(self, name, input, margin, fill_in=0):
    #     return tf.pad(input, margin, constant_values=fill_in, name=name)
    #
    # # -------------------------------------- Arithmetic -------------------------------------- #
    #
    # def clip(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._clip, **kwargs), ['input'], kwargs['name'], [message])
    # def _clip(self, name, input, min_max_vals):
    #     return tf.clip_by_value(input, min_max_vals[0], min_max_vals[1], name)

    def dense(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._dense, **kwargs), ['input'], kwargs['name'], [message])
    def _dense(self, name, input, out_chs, w_init='xavier', w_param=None,
               b_init='zero', b_param=None, w_reg='l2', b_reg='l2'):
        if not w_init is None:
            w_init = self._getInitializer(w_init, w_param, w_reg)
        if not b_init is None:
            use_bias = True
            b_init = self._getInitializer(b_init, b_param, b_reg)
        else:
            use_bias = False
        return nn.Dense(out_chs, use_bias=use_bias, flatten=False, weight_initializer=w_init, bias_initializer=b_init)

    # def argmax(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._argmax, **kwargs), ['input'], kwargs['name'], [message])
    # def _argmax(self, name, input, axis):
    #     return tf.argmax(input, axis, name=name)
    #
    # # ------------------------------------ Image Manipulation ------------------------------------ #
    #
    # def resize(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._resize, **kwargs), ['input'], kwargs['name'], [message])
    # def _resize(self, name, input, size, method='bilinear'):
    #     if method == 'bilinear':
    #         return tf.image.resize_bilinear(input, size, name=name)
    #     elif method == 'bicubic':
    #         return tf.image.resize_bicubic(input, size, name=name)
    #     elif method == 'crop':
    #         return tf.image.resize_image_with_crop_or_pad(input, size, name=name)
    #     else:
    #         raise KeyError('NEBULAE ERROR ⨷ %s is not a legal resize method.' % method)
    #
    #
    # # ------------------------------------ Redefine or Rename ------------------------------------ #
    #
    def duplicate(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._duplicate, **kwargs), ['input'], kwargs['name'], [message])
    def _duplicate(self, name, input):
        return Duplicate()
    #
    # def convert(self, **kwargs):
    #     message = kwargs.get('input', '_INPUT')
    #     return Pod(partial(self._convert, **kwargs), ['input'], kwargs['name'], [message])
    # def _convert(self, name, input, dtype, trainable=False):
    #     '''
    #     convert data type or convert list/numpy array to tensor
    #     :param name:
    #     :param input: input tensor / list / numpy array
    #     :param dtype:
    #     :param trainable: if tensor is trainable
    #     :return: tensor
    #     '''
    #     if isinstance(input, (tf.Tensor, tf.SparseTensor, tf.Variable)):
    #         return tf.cast(input, tf.as_dtype(dtype), name=name)
    #     else:
    #         return tf.Variable(input, trainable=trainable, name=name)

    # ------------------------------------ Loss ------------------------------------ #

    def weight_decay(self, **kwargs):
        return Pod(partial(self._weight_decay, **kwargs), ['penalty'], kwargs['name'],
                   [nd.ones((1)) * kwargs['penalty']])
    def _weight_decay(self, name, penalty, decay_scope=None):
        if not decay_scope is None:
            raise Exception('NEBULAE ERROR ⨷ weight decay cannot be manipulated explicitly in MXNET core.')
        return MXwd()

    def sigm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sigm_xentropy, **kwargs), ['input', 'label'], kwargs['name'], message)
    def _sigm_xentropy(self, name, input, label):
        return MXsigmxe()

    def sftm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sftm_xentropy, **kwargs), ['input', 'label'], kwargs['name'], message)
    def _sftm_xentropy(self, name, input, label, is_one_hot):
        return MXsftmxe(is_one_hot)

    def mse(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mse, **kwargs), ['input', 'label'], kwargs['name'], message)
    def _mse(self, name, input, label):
        return MXmse()

    def mae(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mae, **kwargs), ['input', 'label'], kwargs['name'], message)
    def _mae(self, name, input, label):
        return MXmae()

    # ------------------------------------ Optimizer ------------------------------------ #

    def _lrStrategy(self, lr, lr_decay, miles, param):
        if lr_decay == 'exp':
            return lr_scheduler.FactorScheduler(step=miles, factor=param, base_lr=lr)
        elif lr_decay == 'poly':
            return lr_scheduler.PolyScheduler(max_update=miles, base_lr=lr, pwr=param)
        elif lr_decay == 'cosine':
            pass  #return tf.train.polynomial_decay(lr, mileage, miles, power=param, cycle=True)
        elif lr_decay == 'stair':
            pass  #return tf.train.piecewise_constant(mileage, [i*param for i in range(1, len(lr))], lr)
        else:
            raise KeyError('NEBULAE ERROR ⨷ %s decay is not supported or defined.' % lr_decay)

    def momentum(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._momentum, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
    def _momentum(self, name, lr, input=None, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        optz = Trainer(mod_params, 'SGD',
                       {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    def nesterov(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._nesterov, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
    def _nesterov(self, name, lr, input=None, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):

        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        optz = Trainer(mod_params, 'NAG',
                       {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    def adam(self, **kwargs):
        message = kwargs.get('input', '_INPUT')
        return Pod(partial(self._adam, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
    def _adam(self, name, lr, input=None, mmnt1=0.9, mmnt2=0.999, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt1': mmnt1, 'mmnt2': mmnt2, 'update_scope': update_scope,
                      'ignore_name': ignore_name, 'lr_decay': lr_decay, 'lr_miles': lr_miles,
                      'decay_param': decay_param, 'grad_limit': grad_limit})
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        optz = Trainer(mod_params, 'Adam',
                       {'learning_rate': lr, 'beta1': mmnt1, 'beta2': mmnt2, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    # ------------------------------------ Metric ------------------------------------ #

    def accuracy_cls(self, **kwargs):
        message = []
        message.append(kwargs.get('input', '_INPUT'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._accuracy_cls, **kwargs), ['input', 'label'], kwargs['name'], message)
    def _accuracy_cls(self, name, input, label, multi_class=False, is_one_hot=False):
        return AccuracyCLS(multi_class, is_one_hot)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from math import ceil
import os

from .fuel_tank import FuelTank

class FuelDepot(object):
    def __init__(self):
        self.tanks = {} # contains all fuel tanks
        self.fetcher = {} # contains batch fetcher of all tanks
        self.modifiable_keys = ['name', 'batch_size', 'if_shuffle', 'height', 'width', 'channel', 'resol_ratio',
                                'complete_last_batch', 'spatial_aug', 'p_sa', 'temporal_aug', 'p_ta']

    def loadFuel(self, config=None, name=None, batch_size=None, if_shuffle=True,
                 data_path=None, data_key=None, is_encoded=True, height=0, width=0,
                channel=3, frame=0, rescale=True, resol_ratio=1, complete_last_batch=True,
                 spatial_aug='', p_sa=(0), theta_sa=(0),
                 temporal_aug='', p_ta=(0), theta_ta=(0)):
        if config is None:
            config = {}
            config['name'] = name
            config['batch_size'] = batch_size
            config['data_path'] = data_path
            config['data_key'] = data_key
            config['is_encoded'] = is_encoded
            config['if_shuffle'] = if_shuffle
            config['height'] = height
            config['width'] = width
            config['channel'] = channel
            config['rescale'] = rescale
            config['resol_ratio'] = resol_ratio
            config['complete_last_batch'] = complete_last_batch
            config['frame'] = frame
            config['spatial_aug'] = spatial_aug
            config['p_sa'] = p_sa
            config['theta_sa'] = theta_sa
            config['temporal_aug'] = temporal_aug
            config['p_ta'] = p_ta
            config['theta_ta'] = theta_ta
        else:
            config['if_shuffle'] = config.get('if_shuffle', if_shuffle)
            config['is_encoded'] = config.get('is_encoded', is_encoded)
            config['height'] = config.get('height', height)
            config['width'] = config.get('width', width)
            config['channel'] = config.get('channel', channel)
            config['rescale'] = config.get('rescale', rescale)
            config['resol_ratio'] = config.get('resol_ratio', resol_ratio)
            config['complete_last_batch'] = config.get('complete_last_batch', complete_last_batch)
            config['frame'] = config.get('frame', frame)
            config['spatial_aug'] = config.get('spatial_aug', spatial_aug)
            config['p_sa'] = config.get('p_sa', p_sa)
            config['theta_sa'] = config.get('theta_sa', theta_sa)
            config['temporal_aug'] = config.get('temporal_aug', temporal_aug)
            config['p_ta'] = config.get('p_ta', p_ta)
            config['theta_ta'] = config.get('theta_ta', theta_ta)
        if config['name'] in self.tanks:
            raise Exception('%s is already mounted.' % name)
        if not os.path.exists(config['data_path']):
            raise ValueError('There is no such file or directory named %s.' % config['data_path'])
        self.tanks[config['name']] = FuelTank(config)
        self.fetcher[config['name']] = self.tanks[config['name']]._fetchBatches()

    def unloadFuel(self, tank=''):
        name_err = Exception('NEBULAE ERROR ⨷ %s is not an mounted fuel tank.' % tank)
        if tank != '' and (tank not in self.tanks.keys()):
            raise name_err
        elif tank == '':
            for t in list(self.tanks.keys()):
                self.fetcher.pop(t)
                self.tanks.pop(t)
        else:
            self.fetcher.pop(tank)
            self.tanks.pop(tank)

    def nextBatch(self, name):
        return self.fetcher[name].__next__()

    @property
    def epoch(self):
        epochs = {}
        for t in self.tanks.keys():
            epochs[t] = abs(self.tanks[t].epoch)+1
        return epochs

    @property
    def MPE(self):
        MPEs = {}
        for t in self.tanks.keys():
            MPEs[t] = ceil(self.tanks[t].nsample/self.tanks[t].param['batch_size'])
        return MPEs

    @property
    def volume(self):
        volumes = {}
        for t in self.tanks.keys():
            volumes[t] = self.tanks[t].nsample
        return volumes

    def modify(self, tank='', config=None, **kwargs):
        flag_rename = False
        if config:
            kwargs = config
        for key in kwargs:
            if key not in self.modifiable_keys:
                raise KeyError('NEBULAE ERROR ⨷ %s is not a modifiable parameter or has not been defined.'%key)
            elif key == 'name':
                flag_rename = True
                assert tank!='', 'NEBULAE ERROR ⨷ rename all tanks to same one is illegal.'
                self.tanks[tank].param[key] = kwargs[key]
            else:
                if tank == '':
                    for t in self.tanks.values():
                        t.param[key] = kwargs[key]
                else:
                    self.tanks[tank].param[key] = kwargs[key]
        if flag_rename:
            self.tanks[kwargs['name']] = self.tanks[tank]
            self.tanks.pop(tank)
            self.fetcher[kwargs['name']] = self.fetcher[tank]
            self.fetcher.pop(tank)


#!/usr/bin/env python
'''
space_craft_tf
Created by Seria at 03/02/2019 11:25 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import numpy as np
from .component import Pod


class SpaceCraftTF(object):
    def __init__(self, engine, blueprint, scope=''):
        assert '/' not in scope
        if scope:
            self.scope = scope + '/'
        else:
            self.scope = scope
        self.valid_dtypes = ['uint8', 'int8', 'int16', 'int32', 'int64',
                             'float16', 'float32', 'float64', 'str', 'bool']
        self.operands = {'+': tf.add, '-': tf.subtract, '*': tf.multiply, '@': tf.matmul, '&': tf.concat}
        self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
        self.layout = {'_INPUT': tf.constant(0), '_MEDIA': tf.Variable(0, trainable=False)}
        self.engine = engine
        self.blueprint = blueprint

    def fillUp(self, name, shape, dtype):
        if dtype not in self.valid_dtypes:
            raise Exception('NEBULAE ERROR ⨷ %s is not a valid data type.' % dtype)
        self.layout[name] = tf.placeholder(tf.as_dtype(dtype), shape, name)

    def _getHull(self, component):
        component = self.layout[component]
        if isinstance(component, (tf.Tensor, tf.SparseTensor, tf.Variable)):
            tshape = component.get_shape()
            if tshape.dims is None or tshape.ndims == 0:
                return ' 1    '
            shape = tshape[1:].as_list()
        elif isinstance(component, np.ndarray):
            shape = list(component.shape)
        else:
            shape = [1]
        len_wo_bs = len(shape) - 1
        hull = ' '
        if len_wo_bs < 0:
            return ' 1    '
        for dim in range(len_wo_bs):
            hull += '%-4d x ' % shape[dim]
        hull += '%-5d' % shape[-1]
        return hull

    def _assemble(self, component, assemblage=None, sub_scope=''):
        symbol = component.symbol
        message = component.message
        if sub_scope:
            sub_prefix = sub_scope + '/'
        else:
            sub_prefix = sub_scope
        # this is an atomic component
        if symbol == '>':
            for comp in component.component:
                assemblage = self.assemble(comp, assemblage=assemblage,
                                           sub_scope=sub_scope, is_external=False)
            return assemblage
        elif symbol in ['+', '-', '*', '@', '&']:
            operator = None
            self.operands_counter[symbol] += 1
            # TODO: if a pod is extended to more than two comps, assemblage_name needs modification
            assemblage_name = component.name + '_' + str(self.operands_counter[symbol])
            assemblage_list = []
            hull_list = []
            init_assemblage = assemblage
            for comp in component.component:
                assemblage = self.assemble(comp, assemblage=init_assemblage,
                                           sub_scope=sub_scope, is_external=False)
                assemblage_list.append(assemblage)
                hull = self._getHull(assemblage)
                hull_list.append(hull)
                if operator is None:
                    operator = self.layout[assemblage]
                else:
                    if symbol == '&':
                        operator = self.operands[symbol]([operator, self.layout[assemblage]],
                                                              axis=-1, name=assemblage_name)
                    else:
                        operator = self.operands[symbol](operator,
                                                              self.layout[assemblage],
                                                              name=assemblage_name)
            asmb_full_name = sub_prefix + assemblage_name
            self.layout[asmb_full_name] = operator
            self.blueprint._combineNodes(assemblage_list, hull_list,
                                         asmb_full_name, symbol)
            return asmb_full_name
        else:
            node_name = sub_prefix + component.name
            if node_name in self.layout.keys():  # Reuse or Redefine an existent node
                assert len(message) == 1  # the only way to reuse is invoking DUPLICATE with single input
                hull = self._getHull(message[0])
                self.blueprint._drawNode(message[0], node_name, hull)
                return node_name
            # instantiate input arguments
            kwargs = {'name': node_name}
            for s, sym in enumerate(symbol):
                kwargs[sym] = self.layout[message[s]]

            if assemblage is None:
                self.layout[node_name] = component.component(**kwargs)
                if len(symbol) == 0:
                    self.blueprint._drawNode(node_name, node_name, visible=False)
            else:
                if len(symbol) == 0:
                    self.layout[node_name] = component.component(**kwargs)
                    self.blueprint._drawNode(node_name, node_name, visible=False)
                else:  # message includes input at least
                    kwargs['input'] = self.layout[assemblage]
                    self.layout[node_name] = component.component(**kwargs)
                    # draw layout sheet
                    hull = self._getHull(assemblage)
                    self.blueprint._drawNode(assemblage, node_name, hull)
            for msg in message:
                # draw layout sheet
                if msg[0]!='_' and isinstance(self.layout[msg], (tf.Tensor, tf.SparseTensor, tf.Variable)):
                    hull = self._getHull(msg)
                    self.blueprint._drawNode(msg, node_name, hull)
            return node_name

    def assemble(self, component, sub_scope='', assemblage=None, is_external=True):
        if not isinstance(component, Pod):
            raise TypeError('NEBULAE ERROR ⨷ only the members from Component can be assembled.')
        if not is_external:
            return self._assemble(component, assemblage=assemblage, sub_scope=sub_scope)
        else:
            # append sub scope
            if sub_scope:
                scope = self.scope + sub_scope + '/'
            else:
                scope = self.scope
            # whether to reuse variable name
            reuse = tf.AUTO_REUSE
            with tf.variable_scope(scope, reuse=reuse):
                return self._assemble(component, assemblage=assemblage, sub_scope=sub_scope)


#!/usr/bin/env python
'''
space_craft_mx
Created by Seria at 06/02/2019 8:45 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from mxnet import nd
from mxnet import symbol as syb
from mxnet.gluon import nn, ParameterDict, HybridBlock
import numpy as np
from .component import Pod



class SpaceCraftMX(HybridBlock):
    def __init__(self, engine, blueprint, scope='', **kwargs):
        assert '/' not in scope
        if scope:
            self.scope = scope + '/'
            kwargs['prefix'] = scope
        else:
            self.scope = scope
        super(SpaceCraftMX, self).__init__(**kwargs)
        self.valid_dtypes = ['uint8', 'int8', 'int16', 'int32', 'int64',
                             'float16', 'float32', 'float64', 'str', 'bool']
        self.operands = {'+': nd.add, '-': nd.subtract, '*': nd.multiply, '@': nd.batch_dot, '&': nd.concat}
        self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
        self.layout = {'_INPUT': nd.zeros((1)), '_PENALTY': nd.zeros((1))}
        self.ID = set() # record all assembled components in layout
        self.flag_init = True
        self.debut = True
        self.launcher = {}
        self.terminator = {}
        self.module = []
        self.comp = []
        self.in_node = []
        self.out_node = []
        self.engine = engine
        self.blueprint = blueprint

    def hybrid_forward(self, F, *args):
        for i, k in enumerate(self.in_node):
            self.layout[k] = args[i]
        # reset
        self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
        self.ID = set()
        # parse components
        for mod, scope in self.module:
            _ = self._assemble(mod, sub_scope=scope)
        # only once
        if self.flag_init:
            self.blueprint.log(None)
            self.blueprint.param['verbose'] = False
            # self.hybridize()
            self.flag_init = False
        output = {}
        for node in self.out_node:
            output[node] = self.layout[node]
        return output

    def _getHull(self, component):
        if component not in self.layout:
            shape = self.launcher[component][0]
        else:
            component = self.layout[component]
            if isinstance(component, nd.ndarray.NDArray):
                shape = component.shape[1:] ##########################################################
            elif isinstance(component, np.ndarray):
                shape = list(component.shape)
            else:
                shape = [1]
        # length without batch size
        len_wo_bs = len(shape) - 1
        if len_wo_bs<0:
            return ' 1    '
        hull = ' '
        for dim in range(len_wo_bs):
            hull += '%-4d x ' % shape[dim]
        hull += '%-5d' % shape[-1]
        return hull

    def _assemble(self, component, assemblage=None, sub_scope='', virtual=False):
        '''
        :param component:
        :param assemblage:
        :param sub_scope:
        :param virtual: if True, build up network architecture before hybridization
        :return:
        '''
        symbol = component.symbol
        message = component.message
        if sub_scope:
            sub_prefix = sub_scope + '/'
        else:
            sub_prefix = sub_scope
        if symbol == '>':
            for comp in component.component:
                assemblage = self._assemble(comp, assemblage=assemblage, sub_scope=sub_scope, virtual=virtual)
            return assemblage
        elif symbol in ['+', '-', '*', '@', '&']:
            operator = None
            self.operands_counter[symbol] += 1
            # TODO: if a pod is extended to more than two comps, assemblage_name needs modification
            assemblage_name = component.name + '_' + str(self.operands_counter[symbol])
            assemblage_list = []
            hull_list = []
            init_assemblage = assemblage
            for comp in component.component:
                assemblage = self._assemble(comp, assemblage=init_assemblage, sub_scope=sub_scope, virtual=virtual)
                if virtual:
                    continue
                assemblage_list.append(assemblage)
                hull = self._getHull(assemblage)
                hull_list.append(hull)
                if operator is None:
                    operator = self.layout[assemblage]
                else:
                    # if not self.flag_init:
                    #     if symbol == '&':
                    #         operator = syb.concat([operator, self.layout[assemblage]], dim=-1)
                    #     elif symbol == '@':
                    #         operator = syb.batch_dot(operator, self.layout[assemblage])
                    #     elif symbol == '+':
                    #         operator = operator + self.layout[assemblage]
                    #     elif symbol == '-':
                    #         operator = operator - self.layout[assemblage]
                    #     elif symbol == '*':
                    #         operator = operator * self.layout[assemblage]
                    # else:
                    if symbol == '&':
                        operator = self.operands[symbol]([operator, self.layout[assemblage]], dim=-1)
                    else:
                        operator = self.operands[symbol](operator, self.layout[assemblage])
            asmb_full_name = sub_prefix + assemblage_name
            if virtual:
                return asmb_full_name
            self.layout[asmb_full_name] = operator
            self.ID.add(asmb_full_name)
            self.blueprint._combineNodes(assemblage_list, hull_list,
                                         asmb_full_name, symbol)
            return asmb_full_name
        else: # this is an atomic component
            node_name = sub_prefix + component.name
            # instantiate input arguments
            args = ()
            for s, sym in enumerate(symbol):
                if virtual:
                # special conditions
                    if sym == 'penalty':
                        self.layout['_PENALTY'] = message[s]
                        message[s] = '_PENALTY'
                    elif sym == 'is_train':
                        self.layout['_IS_TRAIN'] = message[s]
                    continue
                elif sym == 'is_train':
                    continue
                args += (self.layout[component.message[s]],)
            if virtual:
                return node_name
            if node_name in self.ID:  # Reuse or Redefine an existent node
                assert len(message) == 1  # the only way to reuse is invoking DUPLICATE with single input
                hull = self._getHull(message[0])
                self.blueprint._drawNode(message[0], node_name, hull)
                return node_name
            if assemblage is None:
                self.layout[node_name] = component.component(*args)
                self.ID.add(node_name)
                if symbol[0] == 'penalty':
                    self.blueprint._drawNode(node_name, node_name, visible=False)
            else:
                if symbol[0] == 'penalty':
                    self.layout[node_name] = component.component(*args)
                    self.ID.add(node_name)
                    self.blueprint._drawNode(node_name, node_name, visible=False)
                else: # message includes input at least
                    args = (self.layout[assemblage],) + args[1:]
                    self.layout[node_name] = component.component(*args)
                    self.ID.add(node_name)
                    # draw layout sheet
                    hull = self._getHull(assemblage)
                    self.blueprint._drawNode(assemblage, node_name, hull)
            for msg in message:
                # draw layout sheet
                if msg[0]!='_' and symbol[0]!='penalty':
                    hull = self._getHull(msg)
                    self.blueprint._drawNode(msg, node_name, hull)
            return node_name

    def _finalize(self, first_call):
        if first_call:
            for c in self.comp:
                self.register_child(c)
            for k in self.terminator.keys():
                self.layout[k] = None
        else:
            if self.debut:
                self.initialize(ctx=self.engine.context)
            for k in self.terminator.keys():
                self.layout[k] = self.terminator[k].component(media=(self.collect_params(), self.layout['_PENALTY']))

    def _allot(self, comp, sub_scope, is_optz):
        if is_optz: # is optimizer
            if sub_scope:
                assemblage = sub_scope + '/' + comp.name
            else:
                assemblage = comp.name
            self.terminator[assemblage] = comp
        else:
            self.module.append((comp, sub_scope))
            self.comp += self.module[-1][0]._build(self.launcher)
            assemblage = self._assemble(comp, sub_scope=sub_scope, virtual=True)
        return assemblage

    def fillUp(self, name, shape, dtype):
        if dtype not in self.valid_dtypes:
            raise Exception('NEBULAE ERROR ⨷ %s is not a valid data type.' % dtype)
        if dtype in ['float16', 'float32', 'float64', 'int32', 'int64', 'int8', 'uint8']:
            self.in_node.append(name)
        self.launcher[name] = (shape, dtype)

    def assemble(self, component, sub_scope=''):
        if not isinstance(component, Pod):
            raise TypeError('NEBULAE ERROR ⨷ only the members in Component can be assembled.')
        optype = component._isOptz()
        if optype == 1:
            assemblage = self._allot(component, sub_scope, True)
        elif optype == -1:
            assemblage = self._allot(component, sub_scope, False)
        else:
            _ = self._allot(component.component[0], sub_scope, False)
            assemblage = self._allot(component.component[1], sub_scope, True)
        return assemblage


#!/usr/bin/env python
'''
navigator
Created by Seria at 23/12/2018 11:21 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ..law import Law

class Stage(object):
    START = 'START'
    RUN = 'RUN'
    END = 'END'
    BREAK_DOWN = 'BREAK_DOWN'
    stages = [START, RUN, END, BREAK_DOWN]
    groups = {START:[], RUN:[], END:[], BREAK_DOWN:[]}
    @classmethod
    def register(cls, group, stage):
        for s,stg in enumerate(stage):
            stg = stg.upper()
            stage[s] = stg
            if stg in cls.stages:
                raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % stage)
            else:
                exec('Stage.%s = "%s"' % (stg, stg))
                cls.stages.append(stg)

        group = group.upper()
        if group not in cls.groups.keys():
            raise KeyError('NEBULAE ERROR ⨷ there is no group named %s.' % group)
        else:
            cls.groups[group].extend(stage)
        # initialize an empty group
        for stg in stage:
            cls.groups[stg] = []



def Navigator(dashboard, time_machine, fuel_depot, spacecraft=None):
    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .navigator_tf import NavigatorTF
        return NavigatorTF(dashboard, time_machine, fuel_depot, spacecraft)
    elif core == 'MXNET':
        from .navigator_mx import NavigatorMX
        return NavigatorMX(dashboard, time_machine, fuel_depot, spacecraft)
    else:
        raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)


# #!/usr/bin/env python
# '''
# space_craft_mx
# Created by Seria at 06/02/2019 8:45 PM
# Email: zzqsummerai@yeah.net
#
#                     _ooOoo_
#                   o888888888o
#                  o88`_ . _`88o
#                  (|  0   0  |)
#                  O \   。   / O
#               _____/`-----‘\_____
#             .’   \||  _ _  ||/   `.
#             |  _ |||   |   ||| _  |
#             |  |  \\       //  |  |
#             |  |    \-----/    |  |
#              \ .\ ___/- -\___ /. /
#          ,--- /   ___\<|>/___   \ ---,
#          | |:    \    \ /    /    :| |
#          `\--\_    -. ___ .-    _/--/‘
#    ===========  \__  NOBUG  __/  ===========
#
# '''
# # -*- coding:utf-8 -*-
# import torch
# import torch.nn as nn
# import numpy as np
# from .component import Pod
#
#
#
# class SpaceCraftPT(nn.Module):
#     def __init__(self, engine, blueprint, scope=''):
#         assert '/' not in scope
#         if scope:
#             self.scope = scope + '/'
#         else:
#             self.scope = scope
#         super(SpaceCraftPT, self).__init__()
#         self.valid_dtypes = ['uint8', 'int8', 'int16', 'int32', 'int64',
#                              'float16', 'float32', 'float64', 'str', 'bool']
#         self.operands = {'+': torch.add, '-': torch.subtract, '*': nd.multiply, '@': nd.batch_dot, '&': nd.concat}
#         self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
#         self.layout = {'_INPUT': nd.zeros((1)), '_PENALTY': nd.zeros((1))}
#         self.ID = set() # record all assembled components in layout
#         self.flag_init = True
#         self.launcher = {}
#         self.terminator = {}
#         self.module = []
#         self.comp = []
#         self.in_node = []
#         self.out_node = []
#         self.engine = engine
#         self.blueprint = blueprint
#
#     def hybrid_forward(self, F, *args):
#         for i, k in enumerate(self.in_node):
#             self.layout[k] = args[i]
#         # reset
#         self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
#         self.ID = set()
#         # parse components
#         for mod, scope in self.module:
#             _ = self._assemble(mod, sub_scope=scope)
#         # only once
#         if self.flag_init:
#             self.blueprint.log(None)
#             self.blueprint.param['verbose'] = False
#             # self.hybridize()
#             self.flag_init = False
#         output = {}
#         for node in self.out_node:
#             output[node] = self.layout[node]
#         return output
#
#     def _getHull(self, component):
#         if component not in self.layout:
#             shape = self.launcher[component][0]
#         else:
#             component = self.layout[component]
#             if isinstance(component, nd.ndarray.NDArray):
#                 shape = component.shape[1:] ##########################################################
#             elif isinstance(component, np.ndarray):
#                 shape = list(component.shape)
#             else:
#                 shape = [1]
#         # length without batch size
#         len_wo_bs = len(shape) - 1
#         if len_wo_bs<0:
#             return ' 1    '
#         hull = ' '
#         for dim in range(len_wo_bs):
#             hull += '%-4d x ' % shape[dim]
#         hull += '%-5d' % shape[-1]
#         return hull
#
#     def _assemble(self, component, assemblage=None, sub_scope='', virtual=False):
#         '''
#         :param component:
#         :param assemblage:
#         :param sub_scope:
#         :param virtual: if True, build up network architecture before hybridization
#         :return:
#         '''
#         symbol = component.symbol
#         message = component.message
#         if sub_scope:
#             sub_prefix = sub_scope + '/'
#         else:
#             sub_prefix = sub_scope
#         if symbol == '>':
#             for comp in component.component:
#                 assemblage = self._assemble(comp, assemblage=assemblage, sub_scope=sub_scope, virtual=virtual)
#             return assemblage
#         elif symbol in ['+', '-', '*', '@', '&']:
#             operator = None
#             self.operands_counter[symbol] += 1
#             # TODO: if a pod is extended to more than two comps, assemblage_name needs modification
#             assemblage_name = component.name + '_' + str(self.operands_counter[symbol])
#             assemblage_list = []
#             hull_list = []
#             init_assemblage = assemblage
#             for comp in component.component:
#                 assemblage = self._assemble(comp, assemblage=init_assemblage, sub_scope=sub_scope, virtual=virtual)
#                 if virtual:
#                     continue
#                 assemblage_list.append(assemblage)
#                 hull = self._getHull(assemblage)
#                 hull_list.append(hull)
#                 if operator is None:
#                     operator = self.layout[assemblage]
#                 else:
#                     # if not self.flag_init:
#                     #     if symbol == '&':
#                     #         operator = syb.concat([operator, self.layout[assemblage]], dim=-1)
#                     #     elif symbol == '@':
#                     #         operator = syb.batch_dot(operator, self.layout[assemblage])
#                     #     elif symbol == '+':
#                     #         operator = operator + self.layout[assemblage]
#                     #     elif symbol == '-':
#                     #         operator = operator - self.layout[assemblage]
#                     #     elif symbol == '*':
#                     #         operator = operator * self.layout[assemblage]
#                     # else:
#                     if symbol == '&':
#                         operator = self.operands[symbol]([operator, self.layout[assemblage]], dim=-1)
#                     else:
#                         operator = self.operands[symbol](operator, self.layout[assemblage])
#             asmb_full_name = sub_prefix + assemblage_name
#             if virtual:
#                 return asmb_full_name
#             self.layout[asmb_full_name] = operator
#             self.ID.add(asmb_full_name)
#             self.blueprint._combineNodes(assemblage_list, hull_list,
#                                          asmb_full_name, symbol)
#             return asmb_full_name
#         else: # this is an atomic component
#             node_name = sub_prefix + component.name
#             # instantiate input arguments
#             args = ()
#             for s, sym in enumerate(symbol):
#                 if virtual:
#                 # special conditions
#                     if sym == 'penalty':
#                         self.layout['_PENALTY'] = message[s]
#                         message[s] = '_PENALTY'
#                     elif sym == 'is_train':
#                         self.layout['_IS_TRAIN'] = message[s]
#                     continue
#                 elif sym == 'is_train':
#                     continue
#                 args += (self.layout[component.message[s]],)
#             if virtual:
#                 return node_name
#             if node_name in self.ID:  # Reuse or Redefine an existent node
#                 assert len(message) == 1  # the only way to reuse is invoking DUPLICATE with single input
#                 hull = self._getHull(message[0])
#                 self.blueprint._drawNode(message[0], node_name, hull)
#                 return node_name
#             if assemblage is None:
#                 self.layout[node_name] = component.component(*args)
#                 self.ID.add(node_name)
#                 if symbol[0] == 'penalty':
#                     self.blueprint._drawNode(node_name, node_name, visible=False)
#             else:
#                 if symbol[0] == 'penalty':
#                     self.layout[node_name] = component.component(*args)
#                     self.ID.add(node_name)
#                     self.blueprint._drawNode(node_name, node_name, visible=False)
#                 else: # message includes input at least
#                     args = (self.layout[assemblage],) + args[1:]
#                     self.layout[node_name] = component.component(*args)
#                     self.ID.add(node_name)
#                     # draw layout sheet
#                     hull = self._getHull(assemblage)
#                     self.blueprint._drawNode(assemblage, node_name, hull)
#             for msg in message:
#                 # draw layout sheet
#                 if msg[0]!='_' and symbol[0]!='penalty':
#                     hull = self._getHull(msg)
#                     self.blueprint._drawNode(msg, node_name, hull)
#             return node_name
#
#     def _finalize(self):
#         for c in self.comp:
#             self.register_child(c)
#         self.initialize(ctx=self.engine.context)
#         for k in self.terminator.keys():
#             self.layout[k] = self.terminator[k].component(media=(self.collect_params(), self.layout['_PENALTY']))
#
#     def _allot(self, comp, sub_scope, is_optz):
#         if is_optz: # is optimizer
#             if sub_scope:
#                 assemblage = sub_scope + '/' + comp.name
#             else:
#                 assemblage = comp.name
#             self.terminator[assemblage] = comp
#         else:
#             self.module.append((comp, sub_scope))
#             self.comp += self.module[-1][0]._build(self.launcher)
#             assemblage = self._assemble(comp, sub_scope=sub_scope, virtual=True)
#         return assemblage
#
#     def fillUp(self, name, shape, dtype):
#         if dtype not in self.valid_dtypes:
#             raise Exception('NEBULAE ERROR ⨷ %s is not a valid data type.' % dtype)
#         if dtype in ['float16', 'float32', 'float64', 'int32', 'int64', 'int8', 'uint8']:
#             self.in_node.append(name)
#         self.launcher[name] = (shape, dtype)
#
#     def assemble(self, component, sub_scope=''):
#         if not isinstance(component, Pod):
#             raise TypeError('NEBULAE ERROR ⨷ only the members in Component can be assembled.')
#         optype = component._isOptz()
#         if optype == 1:
#             assemblage = self._allot(component, sub_scope, True)
#         elif optype == -1:
#             assemblage = self._allot(component, sub_scope, False)
#         else:
#             _ = self._allot(component.component[0], sub_scope, False)
#             assemblage = self._allot(component.component[1], sub_scope, True)
#         return assemblage


# #!/usr/bin/env python
# '''
# component_tf
# Created by Seria at 05/02/2019 1:41 PM
# Email: zzqsummerai@yeah.net
#
#                     _ooOoo_
#                   o888888888o
#                  o88`_ . _`88o
#                  (|  0   0  |)
#                  O \   。   / O
#               _____/`-----‘\_____
#             .’   \||  _ _  ||/   `.
#             |  _ |||   |   ||| _  |
#             |  |  \\       //  |  |
#             |  |    \-----/    |  |
#              \ .\ ___/- -\___ /. /
#          ,--- /   ___\<|>/___   \ ---,
#          | |:    \    \ /    /    :| |
#          `\--\_    -. ___ .-    _/--/‘
#    ===========  \__  NOBUG  __/  ===========
#
# '''
# # -*- coding:utf-8 -*-
# from functools import partial
# from mxnet.gluon import nn, loss, Trainer, HybridBlock
# from mxnet import nd, init, autograd, lr_scheduler
# from math import ceil
# import os
# from ..toolkit.utility import recordConfig
# from .component import Pod
#
# class MXsigmoid(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXsigmoid, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return F.sigmoid(input)
#
# class MXtanh(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXtanh, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return F.tanh(input)
#
# class MXsoftmax(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXsoftmax, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return F.softmax(input)
#
# class MXrelu(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXrelu, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return F.relu(input)
#
# class Duplicate(HybridBlock):
#     def __init__(self, **kwargs):
#         super(Duplicate, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return input
#
# class MXwd(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXwd, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, penalty):
#         return F.zeros((1))
#
# class MXmse(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXmse, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input, label):
#         return (input-label)**2
#
# class MXmae(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXmae, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input, label):
#         return F.abs(input-label)
#
# class MXsigmxe(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MXsigmxe, self).__init__(**kwargs)
#         self.cost = loss.SigmoidBCELoss()
#
#     def hybrid_forward(self, F, input, label):
#         return F.mean(self.cost(input, label))
#
# class MXsftmxe(HybridBlock):
#     def __init__(self, is_one_hot, **kwargs):
#         super(MXsftmxe, self).__init__(**kwargs)
#         self.cost = loss.SoftmaxCELoss(sparse_label=not is_one_hot)
#
#     def hybrid_forward(self, F, input, label):
#         return F.mean(self.cost(input, label))
#
# class MX(HybridBlock):
#     def __init__(self, **kwargs):
#         super(MX, self).__init__(**kwargs)
#
#     def hybrid_forward(self, F, input):
#         return nd.sigmoid(input)
#
# class AccuracyCLS(HybridBlock):
#     def __init__(self, multi_class, is_one_hot, **kwargs):
#         super(AccuracyCLS, self).__init__(**kwargs)
#         if multi_class:
#             assert not is_one_hot
#         self.mulcls = multi_class
#         self.onehot = is_one_hot
#
#     def hybrid_forward(self, F, input, label):
#         if self.mulcls:
#             pred = F.round(input)
#             correct = F.mean(pred == label, axis=-1)
#             return F.mean(correct == 1)
#         else:
#             if self.onehot:
#                 label = F.argmax(label, axis=-1)
#             pred = F.argmax(input, axis=-1)
#             return F.mean(pred == label)
#
# class ComponentMX(object):
#     def __init__(self, channel_major=True, time_major=False):
#         self.channel_major = channel_major
#         self.time_major = time_major
#         self.warehouse = ['CONV_1D', 'CONV', 'CONV_3D', 'CONV_TRANS', 'CONV_SEP',
#                           'SIGMOID', 'TANH', 'SOFTMAX', 'RELU', 'RELU_LEAKY', 'RELU_EXP',
#                           'MAX_POOL', 'AVG_POOL',
#                           'DROPOUT', 'BATCH_NORM', 'EMBEDDING',
#                           'RESHAPE', 'FLAT', 'SLICE',
#                           'CLIP', 'DENSE',
#                           'RESIZE',
#                           'DUPLICATE', 'CONVERT',
#                           'WEIGHT_DECAY', 'SIGM_XENTROPY', 'SFTM_XENTROPY', 'MSE', 'MAE',
#                           'MOMENTUM', 'NESTEROV', 'ADAM']
#         # Convolution
#         self.CONV_1D = 0
#         self.CONV = self.conv_2d
#         self.CONV_3D = 2
#         self.CONV_TRANS = 3
#         self.CONV_SEP = 4
#         # Activation
#         self.SIGMOID = self.sigmoid
#         self.TANH = self.tanh
#         self.SOFTMAX = self.softmax
#         self.RELU = self.relu
#         self.RELU_LEAKY = self.relu_leaky
#         # self.RELU_EXP = self.relu_exp
#         # Pooling
#         self.MAX_POOL = self.max_pool_2d
#         self.AVG_POOL = self.avg_pool_2d
#         # Distributing
#         self.DROPOUT = self.dropout
#         self.BATCH_NORM = self.batch_norm
#         # self.EMBEDDING = self.embedding
#         # Reshape
#         # self.RESHAPE = self.reshape
#         self.FLAT = self.flat
#         # self.SLICE = self.slice
#         # self.PAD = self.pad
#         # Arithmetic
#         # self.CLIP = self.clip
#         self.DENSE = self.dense
#         # self.ARGMAX = self.argmax
#         # Image Manipulation
#         # self.RESIZE = self.resize
#         # Copy or Rename
#         self.DUPLICATE = self.duplicate
#         # self.CONVERT = self.convert
#         # Loss
#         self.WEIGHT_DECAY = self.weight_decay
#         self.SIGM_XENTROPY = self.sigm_xentropy
#         self.SFTM_XENTROPY = self.sftm_xentropy
#         self.MSE = self.mse
#         self.MAE = self.mae
#         # Optimizer
#         self.MOMENTUM = self.momentum
#         self.NESTEROV = self.nesterov
#         self.ADAM = self.adam
#         # Metric
#         self.ACC_CLS = self.accuracy_cls
#
#
#
#     def addComp(self, name, comp, is_complete=False):
#         name = name.upper()
#         if name in self.warehouse:
#             raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)
#         else:
#             self.warehouse.append(name)
#         if is_complete:
#             def customizedComp(name):
#                 return Pod(partial(comp.component, name=name), comp.symbol, name, comp.message)
#         else:
#             assert not isinstance(comp.component, list)
#             def customizedComp(**kwargs):
#                 message = []
#                 # check input so as to update message
#                 for s, sym in enumerate(comp.symbol):
#                     if sym in kwargs:
#                         message.append(kwargs[sym])
#                     else:
#                         message.append(comp.message[s])
#                 return Pod(partial(comp.component, **kwargs), comp.symbol, kwargs['name'], message)
#         exec('self.%s = customizedComp' % name)
#
#     # ------------------------------------ Convolution ------------------------------------ #
#
#     def _getInitializer(self, initializer, param, regularizer):
#         init_err = Exception('NEBULAE ERROR ⨷ %s initializer is not defined or supported.' % initializer)
#         if initializer == 'xavier':
#             if param is None:
#                 param = 'gaussian'
#             else:
#                 if param=='normal':
#                     param = 'gaussian'
#             var = init.Xavier(param)
#         elif initializer == 'uniform':
#             if param is None:
#                 param = 0.5
#             else:
#                 if param[0] != 0:
#                     raise Exception('NEBULAE ERROR ⨷ mean of uniform distributed initializer must be 0 in MXNET core.')
#                 param = param[1]
#             var = init.Uniform(param)
#         elif initializer == 'normal':
#             if param is None:
#                 param = 0.1
#             else:
#                 if param[0]+param[1] != 0:
#                     raise Exception('NEBULAE ERROR ⨷ min and max value of normal distributed initializer must be',
#                                     'opposite numbers in MXNET core.')
#                 param = param[1]
#             var = init.Normal(param)
#         elif initializer == 'zero':
#             var = init.Zero()
#         elif initializer == 'one':
#             var = init.One()
#         else:
#             raise init_err
#
#         if regularizer == 'l2':
#             pass
#         else:
#             raise Exception('NEBULAE ERROR ⨷ %s regularizer is not defined or supported.' % regularizer)
#
#         return var
#
#     def conv_2d(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._conv_2d, **kwargs), ['input'], kwargs['name'], [message])
#     def _conv_2d(self, name, input, kernel, out_chs, w_init='xavier', w_param=None,
#                  b_init=None, b_param=None, w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), keep_size=True):
#         if self.channel_major:
#             format = 'NCHW'
#             h, w = input[0][2:]
#         else:
#             format = 'NHWC'
#             h, w = input[0][1:3]
#         if keep_size:
#             y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] + (dilation[0] - 1) * (kernel[0] - 1) - h) / 2)
#             x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] + (dilation[1] - 1) * (kernel[1] - 1) - w) / 2)
#             padding = (y, x)
#         else:
#             padding = (0, 0)
#         if not w_init is None:
#             w_init = self._getInitializer(w_init, w_param, w_reg)
#         if not b_init is None:
#             use_bias = True
#             b_init = self._getInitializer(b_init, b_param, b_reg)
#         else:
#             use_bias = False
#         return nn.Conv2D(channels=out_chs, kernel_size=kernel, strides=stride, padding=padding, dilation=dilation,
#                          layout=format, use_bias=use_bias, weight_initializer=w_init, bias_initializer=b_init)
#
#     # ------------------------------------ Activation ------------------------------------ #
#
#     def sigmoid(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._sigmoid, **kwargs), ['input'], kwargs['name'], [message])
#     def _sigmoid(self, name, input):
#         return MXsigmoid()
#
#     def tanh(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._tanh, **kwargs), ['input'], kwargs['name'], [message])
#     def _tanh(self, name, input):
#         return MXtanh()
#
#     def softmax(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._tanh, **kwargs), ['input'], kwargs['name'], [message])
#     def _softmax(self, name, input):
#         return MXsoftmax()
#
#     def relu(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._relu, **kwargs), ['input'], kwargs['name'], [message])
#     def _relu(self, name, input):
#         return MXrelu()
#
#     def relu_leaky(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._relu_leaky, **kwargs), ['input'], kwargs['name'], [message])
#     def _relu_leaky(self, name, input, alpha=0.2):
#         return nn.LeakyReLU(alpha, name=name)
#
#     # ------------------------------------ Distributing ------------------------------------ #
#
#     def dropout(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._dropout, **kwargs), ['input'], kwargs['name'], [message])
#     def _dropout(self, name, input, p_drop):
#         return nn.Dropout(p_drop)
#
#     def batch_norm(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('is_train', '_IS_TRAIN'))
#         return Pod(partial(self._batch_norm, **kwargs), ['input', 'is_train'], kwargs['name'], message)
#     def _batch_norm(self, name, input, is_train, beta=False, gamma=False):
#         if self.channel_major:
#             axis = 1
#         else:
#             axis = -1
#         return nn.BatchNorm(axis=axis, momentum=0.9, epsilon=1e-05, center=beta, scale=gamma)
#
#     # def embedding(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._embedding, **kwargs), ['input'], kwargs['name'], [message])
#     # def _embedding(self, name, input, vocabulary, vec_dims, w_init='xavier', w_param=None):
#     #     embd_vec = self._createVar(name+'_vec', [vocabulary, vec_dims], w_init, w_param)
#     #     return tf.nn.embedding_lookup(embd_vec, input, name=name)
#
#     # ------------------------------------ Pooling ------------------------------------ #
#
#     def max_pool_2d(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._max_pool_2d, **kwargs), ['input'], kwargs['name'], [message])
#     def _max_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), keep_size=True, if_global=False):
#         if self.channel_major:
#             format = 'NCHW'
#             h, w = input[0][2:]
#         else:
#             format = 'NHWC'
#             h, w = input[0][1:3]
#         if if_global:
#             kernel = (h, w)
#             padding = (0, 0)
#         else:
#             if keep_size:
#                 y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] - h) / 2)
#                 x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] - w) / 2)
#                 padding = (y, x)
#             else:
#                 padding = (0, 0)
#         return nn.MaxPool2D(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=format)
#
#     def avg_pool_2d(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._avg_pool_2d, **kwargs), ['input'], kwargs['name'], [message])
#     def _avg_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), keep_size=True, if_global=False):
#         if self.channel_major:
#             format = 'NCHW'
#             h, w = input[0][2:]
#         else:
#             format = 'NHWC'
#             h, w = input[0][1:3]
#         if if_global:
#             kernel = (h, w)
#             padding = (0, 0)
#         else:
#             if keep_size:
#                 y = ceil(((ceil(h / stride[0]) - 1) * stride[0] + kernel[0] - h) / 2)
#                 x = ceil(((ceil(w / stride[1]) - 1) * stride[1] + kernel[1] - w) / 2)
#                 padding = (y, x)
#             else:
#                 padding = (0, 0)
#         return nn.AvgPool2D(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=format)
#
#     # ------------------------------------ Reshape ------------------------------------ #
#
#     # def reshape(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._reshape, **kwargs), ['input'], kwargs['name'], [message])
#     # def _reshape(self, name, input, shape):
#     #     return tf.reshape(input, shape, name)
#     #
#     def flat(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._flat, **kwargs), ['input'], kwargs['name'], [message])
#     def _flat(self, name, input):
#         return nn.Flatten()
#     #
#     # def slice(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._slice, **kwargs), ['input'], kwargs['name'], [message])
#     # def _slice(self, name, input, indices):
#     #     for d in range(len(indices)):
#     #         output = input[indices[d][0]:indices[d][1]]
#     #     return tf.identity(output, name)
#     #
#     # def pad(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._pad, **kwargs), ['input'], kwargs['name'], [message])
#     # def _pad(self, name, input, margin, fill_in=0):
#     #     return tf.pad(input, margin, constant_values=fill_in, name=name)
#     #
#     # # -------------------------------------- Arithmetic -------------------------------------- #
#     #
#     # def clip(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._clip, **kwargs), ['input'], kwargs['name'], [message])
#     # def _clip(self, name, input, min_max_vals):
#     #     return tf.clip_by_value(input, min_max_vals[0], min_max_vals[1], name)
#
#     def dense(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._dense, **kwargs), ['input'], kwargs['name'], [message])
#     def _dense(self, name, input, out_chs, w_init='xavier', w_param=None,
#                b_init='zero', b_param=None, w_reg='l2', b_reg='l2'):
#         if not w_init is None:
#             w_init = self._getInitializer(w_init, w_param, w_reg)
#         if not b_init is None:
#             use_bias = True
#             b_init = self._getInitializer(b_init, b_param, b_reg)
#         else:
#             use_bias = False
#         return nn.Dense(out_chs, use_bias=use_bias, flatten=False, weight_initializer=w_init, bias_initializer=b_init)
#
#     # def argmax(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._argmax, **kwargs), ['input'], kwargs['name'], [message])
#     # def _argmax(self, name, input, axis):
#     #     return tf.argmax(input, axis, name=name)
#     #
#     # # ------------------------------------ Image Manipulation ------------------------------------ #
#     #
#     # def resize(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._resize, **kwargs), ['input'], kwargs['name'], [message])
#     # def _resize(self, name, input, size, method='bilinear'):
#     #     if method == 'bilinear':
#     #         return tf.image.resize_bilinear(input, size, name=name)
#     #     elif method == 'bicubic':
#     #         return tf.image.resize_bicubic(input, size, name=name)
#     #     elif method == 'crop':
#     #         return tf.image.resize_image_with_crop_or_pad(input, size, name=name)
#     #     else:
#     #         raise KeyError('NEBULAE ERROR ⨷ %s is not a legal resize method.' % method)
#     #
#     #
#     # # ------------------------------------ Redefine or Rename ------------------------------------ #
#     #
#     def duplicate(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._duplicate, **kwargs), ['input'], kwargs['name'], [message])
#     def _duplicate(self, name, input):
#         return Duplicate()
#     #
#     # def convert(self, **kwargs):
#     #     message = kwargs.get('input', '_INPUT')
#     #     return Pod(partial(self._convert, **kwargs), ['input'], kwargs['name'], [message])
#     # def _convert(self, name, input, dtype, trainable=False):
#     #     '''
#     #     convert data type or convert list/numpy array to tensor
#     #     :param name:
#     #     :param input: input tensor / list / numpy array
#     #     :param dtype:
#     #     :param trainable: if tensor is trainable
#     #     :return: tensor
#     #     '''
#     #     if isinstance(input, (tf.Tensor, tf.SparseTensor, tf.Variable)):
#     #         return tf.cast(input, tf.as_dtype(dtype), name=name)
#     #     else:
#     #         return tf.Variable(input, trainable=trainable, name=name)
#
#     # ------------------------------------ Loss ------------------------------------ #
#
#     def weight_decay(self, **kwargs):
#         return Pod(partial(self._weight_decay, **kwargs), ['penalty'], kwargs['name'],
#                    [nd.ones((1)) * kwargs['penalty']])
#     def _weight_decay(self, name, penalty, decay_scope=None):
#         if not decay_scope is None:
#             raise Exception('NEBULAE ERROR ⨷ weight decay cannot be manipulated explicitly in MXNET core.')
#         return MXwd()
#
#     def sigm_xentropy(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('label', '_LABEL'))
#         return Pod(partial(self._sigm_xentropy, **kwargs), ['input', 'label'], kwargs['name'], message)
#     def _sigm_xentropy(self, name, input, label):
#         return MXsigmxe()
#
#     def sftm_xentropy(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('label', '_LABEL'))
#         return Pod(partial(self._sftm_xentropy, **kwargs), ['input', 'label'], kwargs['name'], message)
#     def _sftm_xentropy(self, name, input, label, is_one_hot):
#         return MXsftmxe(is_one_hot)
#
#     def mse(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('label', '_LABEL'))
#         return Pod(partial(self._mse, **kwargs), ['input', 'label'], kwargs['name'], message)
#     def _mse(self, name, input, label):
#         return MXmse()
#
#     def mae(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('label', '_LABEL'))
#         return Pod(partial(self._mae, **kwargs), ['input', 'label'], kwargs['name'], message)
#     def _mae(self, name, input, label):
#         return MXmae()
#
#     # ------------------------------------ Optimizer ------------------------------------ #
#
#     def _lrStrategy(self, lr, lr_decay, miles, param):
#         if lr_decay == 'exp':
#             return lr_scheduler.FactorScheduler(step=miles, factor=param, base_lr=lr)
#         elif lr_decay == 'poly':
#             return lr_scheduler.PolyScheduler(max_update=miles, base_lr=lr, pwr=param)
#         elif lr_decay == 'cosine':
#             pass  #return tf.train.polynomial_decay(lr, mileage, miles, power=param, cycle=True)
#         elif lr_decay == 'stair':
#             pass  #return tf.train.piecewise_constant(mileage, [i*param for i in range(1, len(lr))], lr)
#         else:
#             raise KeyError('NEBULAE ERROR ⨷ %s decay is not supported or defined.' % lr_decay)
#
#     def momentum(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._momentum, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
#     def _momentum(self, name, lr, input=None, mmnt=0.9, update_scope=None, ignore_name=None,
#                   lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
#         recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
#                      {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
#                       'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
#         # N.B. media is network and wd factor
#         mod_params, wd = media
#         wd = wd.asscalar()
#         if isinstance(lr_decay, str):
#             lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
#         else:
#             lr_s = lr_scheduler.LRScheduler(base_lr=lr)
#         optz = Trainer(mod_params, 'SGD',
#                        {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
#                         'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
#         return optz
#
#     def nesterov(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._nesterov, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
#     def _nesterov(self, name, lr, input=None, mmnt=0.9, update_scope=None, ignore_name=None,
#                   lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
#
#         recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
#                      {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
#                       'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
#         # N.B. media is network and wd factor
#         mod_params, wd = media
#         wd = wd.asscalar()
#         if isinstance(lr_decay, str):
#             lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
#         else:
#             lr_s = lr_scheduler.LRScheduler(base_lr=lr)
#         optz = Trainer(mod_params, 'NAG',
#                        {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
#                         'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
#         return optz
#
#     def adam(self, **kwargs):
#         message = kwargs.get('input', '_INPUT')
#         return Pod(partial(self._adam, **kwargs), ['input', 'media'], kwargs['name'], [message, '_MEDIA'])
#     def _adam(self, name, lr, input=None, mmnt1=0.9, mmnt2=0.999, update_scope=None, ignore_name=None,
#                   lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
#         recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
#                      {'lr': lr, 'mmnt1': mmnt1, 'mmnt2': mmnt2, 'update_scope': update_scope,
#                       'ignore_name': ignore_name, 'lr_decay': lr_decay, 'lr_miles': lr_miles,
#                       'decay_param': decay_param, 'grad_limit': grad_limit})
#         # N.B. media is network and wd factor
#         mod_params, wd = media
#         wd = wd.asscalar()
#         if isinstance(lr_decay, str):
#             lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
#         else:
#             lr_s = lr_scheduler.LRScheduler(base_lr=lr)
#         optz = Trainer(mod_params, 'Adam',
#                        {'learning_rate': lr, 'beta1': mmnt1, 'beta2': mmnt2, 'wd': wd,
#                         'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
#         return optz
#
#     # ------------------------------------ Metric ------------------------------------ #
#
#     def accuracy_cls(self, **kwargs):
#         message = []
#         message.append(kwargs.get('input', '_INPUT'))
#         message.append(kwargs.get('label', '_LABEL'))
#         return Pod(partial(self._accuracy_cls, **kwargs), ['input', 'label'], kwargs['name'], message)
#     def _accuracy_cls(self, name, input, label, multi_class=False, is_one_hot=False):
#         return AccuracyCLS(multi_class, is_one_hot)


#!/usr/bin/env python
'''
time_machine_tf
Created by Seria at 04/02/2019 4:35 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from mxnet.gluon import nn, SymbolBlock, HybridBlock
import os
from glob import glob


class TimeMachineMX(object):
    def __init__(self, param):
        '''
        Time Machine saves current states or restores saved states
        '''
        self.param = param



    def _setParams(self, spacecraft):
        self.counter = 0
        self.sc = spacecraft

    def backTo(self, frozen=False, ins=None, outs=None):
        if self.param['ckpt_path'] is None:
            raise Exception('NEBULAE ERROR ⨷ anchor location is not provided.')
        else:
            # spot = SymbolBlock.imports(self.param['ckpt_path'] + '.json', self.sc.in_node,
            #                            self.param['ckpt_path'] + '.params')
            architecture = glob(os.path.join(self.param['ckpt_path'],'*.params'))
            max_saved = -1
            for arch in architecture:
                saved_no = int(arch.split('.')[0].split('-')[-1])
                if saved_no > max_saved:
                    moment = arch
                    max_saved = saved_no
            if frozen:
                self.sc.load_parameters(moment, self.sc.engine.context)
            else:
                missing = True
                extra = True
                self.sc.load_parameters(moment, self.sc.engine.context, allow_missing=missing, ignore_extra=extra)
            print('+' + ((10 + len(self.param['ckpt_path'])) * '-') + '+')
            print('| Back to \033[1;34m%s\033[0m |' % self.param['ckpt_path'])
            print('+' + ((10 + len(self.param['ckpt_path'])) * '-') + '+')

    def dropAnchor(self, frozen=False, anchor=None):
        if self.param['save_path'] is None:
            raise Exception('NEBULAE ERROR ⨷ there is nowhere to drop anchor.')
        else:
            if self.sc.scope:
                scope = self.sc.scope[:-1]
            else:
                scope = 'ckpt'
            if frozen:
                self.sc.export(os.path.join(self.param['save_path'], scope), epoch=self.counter)
            self.sc.save_parameters(os.path.join(self.param['save_path'], '%s-%d.params'%(scope, self.counter)))
            self.counter += 1
            print('| Anchor is dropped at \033[1;34m%s\033[0m |' % self.param['save_path'])


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import math
import io
import random as rand
import numpy as np
import h5py
from PIL import Image
import skimage.transform as imgtrans
from multiprocessing import Manager, Process, cpu_count, Pool

from ..toolkit.decorator import Timer

# def _preprocess(param, queue):
#     atr, curr_sample = queue.get()
#     if atr == param['data_key']:
#         if param['frame'] > 0:  # if this is sequential data
#             temp_data = np.zeros((param['frame'], param['height'],
#                                   param['width'], param['channel']))
#             for f in range(curr_sample.shape[0]):
#                 if f < param['frame']:
#                     temp_data[f] = _byte2Array(param, curr_sample[f])
#                 else:
#                     break
#             curr_data = _dataAugmentation(param, temp_data)
#         else:
#             curr_data = _dataAugmentation(param, _byte2Array(param, curr_sample))
#     else:
#         curr_data = curr_sample
#     return atr, curr_data
#
# def _dataAugmentation(param, source):
#     def _spatialAugmentation(img, da_type, theta):
#         data_aug_err = Exception('%s spatial augmentation is unsupported.' % da_type)
#         if da_type == 'flip':
#             aug_img = np.zeros(img.shape)
#             if param['channel'] == 1:
#                 aug_img = np.fliplr(img)
#             else:
#                 for c in range(param['channel']):
#                     aug_img[:, :, c] = np.fliplr(img[:, :, c])
#             return aug_img
#         elif da_type == 'crop':
#             # theta: ((area_ratio_inf, area_ratio_sup), (aspect_ratio_inf, aspect_ratio_sup))
#             sqrt_aspect_ratio = math.sqrt(rand.random() * (theta[1][1] - theta[1][0]) + theta[1][0])
#             sqrt_area_ratio = math.sqrt(rand.random() * (theta[0][1] - theta[0][0]) + theta[0][0])
#             w = int(sqrt_area_ratio * param['width'] * sqrt_aspect_ratio)
#             h = int(sqrt_area_ratio * param['height'] / sqrt_aspect_ratio)
#             x = int(rand.random() * (param['width'] - w + 1))
#             y = int(rand.random() * (param['height'] - h + 1))
#             # import pdb
#             # pdb.set_trace()
#             aug_img = imgtrans.resize(img[y:y + h, x:x + w].copy(), (param['height'], param['width']))
#             return aug_img
#         elif da_type == 'brightness':
#             perturbation = rand.random() * (theta[1] - theta[0]) + theta[0]
#             aug_img = perturbation + img  # pixels range between [theta, 1+theta]
#             aug_img = np.clip(aug_img, 0, 1)  # pixels range between [0, 1]
#             return aug_img
#         elif da_type == 'gamma_contrast':
#             perturbation = rand.random() * (theta[1] - theta[0]) + theta[0]
#             aug_img = perturbation * (img - 0.5)  # pixels range between [0, theta]
#             aug_img = np.clip(aug_img, -0.5, 0.5) + 0.5  # pixels range between [0, 1]
#             return aug_img
#         elif da_type == 'log_contrast':
#             perturbation = rand.random() * (theta[1] - theta[0]) + theta[0]
#             aug_img = perturbation * np.log(img + 1)  # pixels range between [0, theta * ln2]
#             aug_img = np.where(aug_img > 1, 1, aug_img)  # pixels range between [0, 1]
#             return aug_img
#         else:
#             raise data_aug_err
#
#     def _temporalAugmentation(seq, da_type, theta):
#         data_aug_err = Exception('%s temporal augmentation is unsupported.' % da_type)
#         if da_type == 'sample':
#             seq_len = param['frame'] // theta  # here theta is sampling frequency
#             aug_seq = np.zeros((seq_len,) + seq.shape[1:])
#             for i in range(seq_len):
#                 aug_seq[i] = seq[int((i + rand.random()) * theta)]
#             return aug_seq
#         else:
#             raise data_aug_err
#
#     if param['frame'] != 0:  # if this call is for sequential data augmentation
#         for t, ta in enumerate(param['temporal_aug'].split(',')):
#             if not ta:
#                 continue
#             if rand.random() < param['p_ta'][t]:
#                 source = _temporalAugmentation(source, ta, param['theta_ta'][t])
#                 for s, sa in enumerate(param['spatial_aug'].split(',')):
#                     if not sa:
#                         continue
#                     if rand.random() < param['p_sa'][s]:
#                         for i in range(source.shape[0]):
#                             source[i] = _spatialAugmentation(source[i], sa, param['theta_sa'][t])
#     else:
#         for s, sa in enumerate(param['spatial_aug'].split(',')):
#             if not sa:
#                 continue
#             if rand.random() < param['p_sa'][s]:
#                 source = _spatialAugmentation(source, sa, param['theta_sa'][s])
#
#     return source * 2 - 1  # pixels range between [-1, 1]
#
#
# def _byte2Array(param, data_src):
#     if not param['is_encoded']:  # no need to recover data from bytes
#         return data_src
#     data_bytes = data_src.tobytes()
#     data_pil = Image.open(io.BytesIO(data_bytes))
#     if param['width'] > 0 and param['height'] > 0:
#         # if need lower the image resolution
#         if param['resol_ratio'] < 1:
#             data_pil = data_pil.resize((int(param['resol_ratio'] * param['width']),
#                                         int(param['resol_ratio'] * param['height'])))
#         data_pil = data_pil.resize((param['width'], param['height']))
#     data_np = np.array(data_pil)  # pixels range between [0, 255]
#     data_np = (data_np / 255).astype(np.float32)  # pixels range between [0, 1]
#     if param['channel'] == 1:
#         data_np = np.expand_dims(data_np, -1)
#     return data_np

class FuelTank(object):
    def __init__(self, param):
        self.param = param
        self.epoch = 0
        duration = self._loadData()
        print('+' + (61 * '-') + '+')
        print('| \033[1;35m%-17s\033[0m fuel tank has been mounted within \033[1;35m%6.3fs\033[0m |'
              % (param['name'], duration))
        print('+' + (61 * '-') + '+')

    def __del__(self):
        self.datafile.close()
        print('+' + (50 * '-') + '+')
        print('| \033[1;35m%-17s\033[0m fuel tank is no longer mounted |' % self.param['name'])
        print('+' + (50 * '-') + '+')

    def _shuffleData(self):
        '''
        Args:
        num: batch size
        leftover: the rest samples of last epoch
        '''
        nsample = self.nsample
        for j in range(nsample):
            idx = int(rand.random() * (nsample - j)) + j
            temp = self.order[j]
            self.order[j] = self.order[idx]
            self.order[idx] = temp

    def _dataAugmentation(self, source):
        def _spatialAugmentation(img, da_type, theta):
            data_aug_err = Exception('NEBULAE ERROR ⨷ %s spatial augmentation is unsupported.' % da_type)
            if da_type == 'flip':
                aug_img = np.zeros(img.shape)
                if self.param['channel'] == 1:
                    aug_img = np.fliplr(img)
                else:
                    for c in range(self.param['channel']):
                        aug_img[:, :, c] = np.fliplr(img[:, :, c])
                return aug_img
            elif da_type == 'crop':
                # theta: ((area_ratio_inf, area_ratio_sup), (aspect_ratio_inf, aspect_ratio_sup))
                sqrt_aspect_ratio = math.sqrt(rand.random() * (theta[1][1]-theta[1][0]) + theta[1][0])
                sqrt_area_ratio = math.sqrt(rand.random() * (theta[0][1]-theta[0][0]) + theta[0][0])
                w = int(sqrt_area_ratio*self.param['width']*sqrt_aspect_ratio)
                h = int(sqrt_area_ratio*self.param['height']/sqrt_aspect_ratio)
                x = int(rand.random() * (self.param['width']-w+1))
                y = int(rand.random() * (self.param['height']-h+1))
                # import pdb
                # pdb.set_trace()
                aug_img = imgtrans.resize(img[y:y+h, x:x+w].copy(), (self.param['height'], self.param['width']))
                return aug_img
            elif da_type == 'brightness':
                perturbation = rand.random() * theta * 2 - theta
                aug_img = perturbation + img  # pixels range between [theta, 1+theta]
                aug_img = np.clip(aug_img, 0, 1)  # pixels range between [0, 1]
                return aug_img
            elif da_type == 'gamma_contrast':
                perturbation = rand.random() * theta * 2 - theta
                aug_img = perturbation * (img - 0.5) # pixels range between [0, theta]
                aug_img = np.clip(aug_img, -0.5, 0.5) + 0.5 # pixels range between [0, 1]
                return aug_img
            elif da_type == 'log_contrast':
                perturbation = rand.random() * theta * 2 - theta
                aug_img = perturbation * np.log(img + 1) # pixels range between [0, theta * ln2]
                aug_img = np.where(aug_img > 1, 1, aug_img)  # pixels range between [0, 1]
                return aug_img
            else:
                raise data_aug_err

        def _temporalAugmentation(seq, da_type, theta):
            data_aug_err = Exception('NEBULAE ERROR ⨷ %s temporal augmentation is unsupported.' % da_type)
            if da_type == 'sampling':
                seq_len = self.param['frame'] // theta # here theta is sampling frequency
                aug_seq = np.zeros((seq_len,) + seq.shape[1:])
                for i in range(seq_len):
                    aug_seq[i] = seq[int((i + rand.random()) * theta)]
                return aug_seq
            else:
                raise data_aug_err

        if self.param['frame'] != 0: # if this call is for sequential data augmentation
            for t,ta in enumerate(self.param['temporal_aug'].split(',')):
                if not ta:
                    continue
                if rand.random() < self.param['p_ta'][t]:
                    source = _temporalAugmentation(source, ta, self.param['theta_ta'][t])
                    for s, sa in enumerate(self.param['spatial_aug'].split(',')):
                        if not sa:
                            continue
                        if rand.random() < self.param['p_sa'][s]:
                            for i in range(source.shape[0]):
                                source[i] = _spatialAugmentation(source[i], sa, self.param['theta_sa'][t])
        else:
            for s,sa in enumerate(self.param['spatial_aug'].split(',')):
                if not sa:
                    continue
                if rand.random() < self.param['p_sa'][s]:
                    source = _spatialAugmentation(source, sa, self.param['theta_sa'][s])

        if self.param['rescale']:
            return source * 2 - 1 # pixels range between [-1, 1]
        else:
            return source

    @Timer
    def _loadData(self):
        # --------------------------------- load data --------------------------------- #
        format = self.param['data_path'].split('.')[-1]
        if format == 'npz':
            self.datafile = np.load(self.param['data_path'])

        elif format == 'hdf5':
            self.datafile = h5py.File(self.param['data_path'], 'r')


        self.attributes = [key for key in self.datafile.keys()]
        self.dtypqeses = {}
        if self.param['frame'] == -1:
            self.param['frame'] = self.datafile['_MAX_FRAMES'][()]
            self.attributes.remove('_MAX_FRAMES')
        self.dtypes = {}
        for atr in self.attributes:
            if atr == self.param['data_key']:
                self.dtypes[atr] = np.dtype('float32')
            else:
                self.dtypes[atr] = self.datafile[atr].dtype

        # get number of samples and channels
        self.nsample = self.datafile[self.param['data_key']].shape[0]
        if self.param['batch_size'] > self.nsample:
            raise ValueError('NEBULAE ERROR ⨷ batch size is greater than total number of samples.')

        # --------------------------- initialize and shuffle --------------------------- #
        order = [i for i in range(self.nsample)]
        self.curr_idx = 0
        self.order = order

    def _byte2Array(self, data_src):
        if not self.param['is_encoded']: # no need to recover data from bytes
            return data_src
        data_bytes = data_src.tobytes()
        data_pil = Image.open(io.BytesIO(data_bytes))
        if self.param['width']>0 and self.param['height']>0:
            # if need lower the image resolution
            if self.param['resol_ratio'] < 1:
                data_pil = data_pil.resize((int(self.param['resol_ratio'] * self.param['width']),
                                            int(self.param['resol_ratio'] * self.param['height'])))
            data_pil = data_pil.resize((self.param['width'], self.param['height']))
        data_np = np.array(data_pil) # pixels range between [0, 255]
        data_np = (data_np/255).astype(np.float32) # pixels range between [0, 1]
        if self.param['channel'] == 1:
            data_np = np.expand_dims(data_np, -1)
        return data_np

    def _fetchBatches(self):
        '''
        Fetch next batch
        '''
        while 1:
            if self.curr_idx==0:
                if self.param['if_shuffle']:
                    # reshuffle data to read
                    self._shuffleData()
                else:
                    self.order = [i for i in range(self.nsample)]
            self.curr_idx += self.param['batch_size']
            curr_batch = {}
            for atr in self.attributes:
                curr_batch[atr] = []
            # start a new epoch
            low_power = False
            indices = []
            if self.curr_idx > self.nsample:
                low_power = True
                self.epoch = -self.epoch
                # if the samples left are not enough to feed a mini batch, let's say n samples.
                # the current batch should be the last n samples in addition to the first (bs-n) samples
                indices.extend(self.order[self.curr_idx - self.param['batch_size']:self.nsample])
                if self.param['complete_last_batch']:
                    indices.extend(self.order[0:self.curr_idx - self.nsample])
                self.curr_idx = 0
            else:
                indices.extend(self.order[self.curr_idx - self.param['batch_size']:self.curr_idx])

            for idx in indices:
                for atr in self.attributes:
                    curr_sample = self.datafile[atr][idx]
                    if atr == self.param['data_key']:
                        if self.param['frame'] > 0:  # if this is sequential data
                            if self.param['height'] > 0 and self.param['width'] > 0:
                                temp_data = np.zeros((self.param['frame'], self.param['height'],
                                                      self.param['width'], self.param['channel']))
                            else:
                                temp_data = np.zeros((self.param['frame'], self.param['channel']))
                            for f in range(curr_sample.shape[0]):
                                if f < self.param['frame']:
                                    temp_data[f] = self._byte2Array(curr_sample[f])
                                else:
                                    break
                            curr_batch[atr] += [self._dataAugmentation(temp_data)]
                        else:
                            curr_batch[atr] += [self._dataAugmentation(self._byte2Array(curr_sample))]
                    else:
                        curr_batch[atr] += [curr_sample]



            # cores = cpu_count() // 1
            # print("Creating %d-process pool" % cores)
            # pool = Pool(cores)
            # with Manager() as m:
            #     queue = m.Queue()
            #     for idx in indices:
            #         for atr in self.attributes:
            #             queue.put((atr, self.datafile[atr][idx]))
            #     param = m.dict(self.param)
            #     while not queue.empty():
            #         atr, curr_data = pool.apply(_preprocess, args=(param, queue))
            #         curr_batch[atr].append(curr_data)
            # pool.close()
            # pool.join()



            for atr in self.attributes:
                curr_batch[atr] = np.array(curr_batch[atr]).astype(self.dtypes[atr])
            yield curr_batch
            if low_power:
                # epoch counter turns positive when an epoch ends
                self.epoch = abs(self.epoch) + 1


#!/usr/bin/env python
'''
navigator_tf
Created by Seria at 04/02/2019 4:48 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from mxnet import nd, autograd
import numpy as np
import time
from copy import deepcopy
from collections import defaultdict
import os
from ..toolkit import parseConfig, recordConfig
from .navigator import Stage



class NavigatorMX(object):
    roman_numeral = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X']
    def __init__(self, dashboard, time_machine, fuel_depot, spacecraft):
        self.db = dashboard
        self.tm = time_machine
        self.fd = fuel_depot
        self.sc = spacecraft
        if spacecraft is None:
            self.flag_empty = True
        else:
            self.flag_empty = False
        self.propellants = defaultdict(list)
        self.epoch = 0
        self.mile = 0
        self.mpe = 0
        self.stage = 'UNKNOWN'
        self.feed(Stage.START, time_machine._setParams, external=False, spacecraft=spacecraft)
        self.feed(Stage.START, spacecraft._finalize, external=False, first_call=True) # finalize the architecture
        self.feed(Stage.RUN, self.sc._finalize, external=False, first_call=False)

    def feed(self, stage, propellant, position=None, external=True, **kwargs):
        if external:
            kwargs['nvg'] = self
        if isinstance(position, int):
            quantity = len(self.propellants[stage])
            if position < quantity:
                self.propellants[stage][position] = (propellant, kwargs)
            else:
                for p in range(quantity, position):
                    self.propellants[stage].append((None, 'False'))
        self.propellants[stage].append((propellant, kwargs))

    def _airDrop(self, stage, n, **kwargs):
        # kwargs.update(self.propellants[stage][n][2])
        self.propellants[stage][n] = self.propellants[stage][n][0:2] + (kwargs,)

    def _triggerStage(self, curr_stage):
        self.stage = curr_stage
        for ppl, kwargs in self.propellants[curr_stage]:
            ppl(**kwargs)

    def cruise(self, exit, feedback=None, thruster=None, **entrance):
        '''
        exit: which node to output
        feedback: loss name
        thruster: optimizer name
        entrance: input (formated as keyword arguments)
        '''
        ent = []
        bs = -1
        for k in entrance.keys():
            if isinstance(entrance[k], np.ndarray):
                ent.append(nd.array(entrance[k]).as_in_context(self.sc.engine.context))
                if bs < 0:
                    bs = entrance[k].shape[0]

        if not(feedback is None or feedback in exit):
            exit = exit + (feedback,)

        if thruster is None:
            self.sc.out_node = exit
            with autograd.predict_mode():
                ext = self.sc(*tuple(ent))
        else:
            self.sc.out_node = exit + (thruster,)
            with autograd.record():
                ext = self.sc(*tuple(ent))
            ext[feedback].backward()
            ext[thruster].step(bs)

        ret = {}
        for key in exit:
            if ext[key].size == 1:
                ret[key] = ext[key].asscalar()
            else:
                ret[key] = ext[key].asnumpy()
        return ret

    def execute(self):
        root_stage = self.stage
        for stg in Stage.groups[root_stage]:
            self._triggerStage(stg)
        self.stage = root_stage

    def _scribe(self):
        if self.flag_empty:
            bp_cfg = {}
        else:
            bp_cfg = self.sc.blueprint.param
        config = {'NG': self.sc.engine.param, 'TM': self.tm.param,
                  'FD': {}, 'DB': self.db.param, 'BP': bp_cfg}
        ft = self.fd.tanks
        for k in ft.keys():
            config['FD'][k] = ft[k].param

        temp_config = os.path.join(os.getcwd(), 'temp_config.json')
        if os.path.exists(temp_config):
            hyper_param = parseConfig(temp_config)
            config['SC'] = hyper_param
            if not os.path.exists(self.tm.param['save_path']):
                os.mkdir(self.tm.param['save_path'])
            recordConfig(os.path.join(self.tm.param['save_path'], 'config.json'), config)
            os.remove(os.path.join(os.getcwd(), 'temp_config.json'))

    def log(self, gauge=True, tachograph=True, blueprint=True, setting=True):
        self.feed(Stage.END, self.db._log, external=False, gauge=gauge, tachograph=tachograph)

        if blueprint:
            self.sc.blueprint._await = True
            self.sc.blueprint._path = self.db.param['log_path']
        if setting:
            self.feed(Stage.START, self._scribe, external=False)

    def launch(self):
        try:
            print('+' + 35 * '-' + '+')
            print('| The Spacecraft is being launched. |')
            print('+' + 35 * '-' + '+')
            self._triggerStage(Stage.START)
            self._triggerStage(Stage.RUN)
            self._triggerStage(Stage.END)
            print('+' + 44 * '-' + '+')
            print('| The Spacecraft has arrived at destination. |')
            print('+' + 44 * '-' + '+')

        except BaseException as e:
            if Stage.BREAK_DOWN in self.propellants:
                self._triggerStage(Stage.BREAK_DOWN)
            else:
                raise e


#!/usr/bin/env python
'''
navigator_tf
Created by Seria at 04/02/2019 4:48 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import time
from copy import deepcopy
from collections import defaultdict
import os
from ..toolkit import parseConfig, recordConfig
from .navigator import Stage

# class NavigatorTF(object):
#     roman_numeral = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X']
#     def __init__(self, dashboard, time_machine, fuel_depot, spacecraft):
#         config = {'NG': spacecraft.engine.param, 'TM':time_machine.param,
#                   'FD': {}, 'DB': dashboard.param, 'BP': spacecraft.blueprint.param}
#         ft = fuel_depot.tanks
#         for k in ft.keys():
#             config['FD'][k] = ft[k].param
#
#         self.propellants = defaultdict(list)
#         self.epoch = 0
#         self.mile = 0
#         self.mpe = 0
#         self.batch = {}
#         self.destination = {}
#         self.macro_stage = ''
#         # start a session
#         self.sess = tf.Session(config=spacecraft.engine.config_proto)
#         self.time_machine = time_machine
#         self.fuel_depot = fuel_depot
#         self.spacecraft = spacecraft
#         self.dashboard = dashboard
#         self.feed(Stage.START, self.time_machine._setParams,
#                             sess=self.sess, mile=self.spacecraft.layout['_MEDIA'],
#                             scope=self.spacecraft.scope, config=config)
#         self.feed(Stage.START, self.time_machine.backTo)
#
#     def feed(self, stage, propellant, position=None, condition='True', **kwargs):
#         if isinstance(position, int):
#             quantity = len(self.propellants[stage])
#             if position < quantity:
#                 self.propellants[stage][position] = (propellant, condition, kwargs)
#             else:
#                 for p in range(quantity, position):
#                     self.propellants[stage].append((None, 'False'))
#         self.propellants[stage].append((propellant, condition, kwargs))
#
#     def _airDrop(self, stage, n, **kwargs):
#         # kwargs.update(self.propellants[stage][n][2])
#         self.propellants[stage][n] = self.propellants[stage][n][0:2] + (kwargs,)
#
#     def _triggerStage(self, curr_stage):
#         for stg in Stage.stages:
#             exec('%s = "%s"' % (stg, stg))
#         mile = self.mile+1
#         epoch = self.epoch+1
#         stage = self.macro_stage
#         stats = []
#         cnt = 0
#         for ppl, cond, kwargs in self.propellants[curr_stage]:
#             cnt += 1
#             if_meet = eval(cond)
#             if (ppl is None) or (not if_meet):
#                 stats.append(None)
#                 continue
#             stats.append(ppl(**kwargs))
#         return stats
#
#     def runAMile(self, fuel_line, pass_by=(), destination=(), interval=1):
#         assert len(pass_by)+len(destination) > 0
#         start = time.time()
#         fl = {} # feed dict
#         filled_fl = {}
#         quan_preflight = len(self.propellants[Stage.PREFLIGHT])
#
#         for k in fuel_line.keys():
#             if isinstance(fuel_line[k], str):
#                 pipe_out = self.batch[fuel_line[k]]
#             else:
#                 pipe_out = fuel_line[k]
#             fl[self.spacecraft.layout[k]] = pipe_out
#             filled_fl[k] = pipe_out
#         if quan_preflight > 0:  # there is propellant in Preflight stage
#             for i in range(quan_preflight):
#                 self._airDrop(Stage.PREFLIGHT, i, fuel_line=filled_fl)
#             temp_fl = self._triggerStage(Stage.PREFLIGHT)
#             for tfl in temp_fl:
#                 if not tfl is None:
#                     for k in tfl.keys():
#                         fl[self.spacecraft.layout[k]] = tfl[k]
#                     break
#
#         pb = [self.spacecraft.layout[k] for k in pass_by]
#         dst = [self.spacecraft.layout[k] for k in destination]
#         spot = self.sess.run(pb + dst, fl)
#         spot = spot[len(pass_by):len(pass_by)+len(destination)]
#         sp = {}
#         filled_sp = {}
#         for j in range(len(destination)):
#             filled_sp[destination[j]] = spot[j]
#         quan_analyzer = len(self.propellants[Stage.ANALYZE])
#         if quan_analyzer == 0:  # no propellants in Analyze stage
#             sp = deepcopy(filled_sp)
#         else:
#             for j in range(quan_analyzer):
#                 self._airDrop(Stage.ANALYZE, j, fuel_line=filled_fl, destination=filled_sp)
#             temp_sp = self._triggerStage(Stage.ANALYZE)
#             for tsp in temp_sp:
#                 if not tsp is None:
#                     sp = tsp
#                     break
#
#         prefix = self.macro_stage + ':'
#         for k in sp.keys():
#             self.destination[prefix + k] = sp[k]
#         self.dashboard._gauge(self.destination.items(), self.mile, self.epoch, self.mpe,
#                                   time.time()-start, interval)
#
#     def runAnEpoch(self, fuel_tank, miles=-1):
#         self.mpe = self.fuel_depot.MPE[fuel_tank]
#         if miles < 0:
#             miles = self.mpe
#         # cores = cpu_count() // 1
#         # print("Creating %d-process pool" % cores)
#         # pool = Pool(cores)
#         for mi in range(miles):
#             self.mile = mi
#             # begin = time.time()
#             self.batch = self.fuel_depot.nextBatch(fuel_tank)
#             # print('%.3fs' % (time.time() - begin))
#             for stg in Stage.phases[-1]:
#                 self._triggerStage(stg)
#
#         self.destination = {}
#
#     def log(self, guage=True, tachograph=True, blueprint=True):
#         self.dashboard._log(guage, tachograph)
#         if blueprint:
#             self.spacecraft.blueprint.log(self.dashboard.param['log_path'])
#
#     def launch(self, epochs=1):
#         try:
#             print('+' + 35 * '-' + '+')
#             print('| The Spacecraft is being launched. |')
#             print('+' + 35 * '-' + '+')
#             self._triggerStage(Stage.START)
#             if epochs < 0 or not isinstance(epochs, int):
#                 raise ValueError('NEBULAE ERROR ⨷ epochs must be a non-positive integer.')
#             for ep in range(epochs):
#                 self.epoch = ep
#                 for stg in Stage.phases[0]:
#                     self.macro_stage = stg
#                     self._triggerStage(stg)
#             self._triggerStage(Stage.END)
#             print('+' + 44 * '-' + '+')
#             print('| The Spacecraft has arrived at destination. |')
#             print('+' + 44 * '-' + '+')
#
#         except BaseException as e:
#             if Stage.BREAK_DOWN in self.propellants:
#                 self._triggerStage(Stage.BREAK_DOWN)
#             else:
#                 raise e















class NavigatorTF(object):
    roman_numeral = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X']
    def __init__(self, dashboard, time_machine, fuel_depot, spacecraft):
        self.db = dashboard
        self.tm = time_machine
        self.fd = fuel_depot
        self.sc = spacecraft
        if spacecraft is None:
            self.flag_empty = True
        else:
            self.flag_empty = False
        self.propellants = defaultdict(list)
        self.epoch = 0
        self.mile = 0
        self.mpe = 0
        self.stage = 'UNKNOWN'
        # start a session
        self.sess = tf.Session(config=self.sc.engine.config_proto)
        self.feed(Stage.START, time_machine._setParams, external=False,
                            sess=self.sess, mile=spacecraft.layout['_MEDIA'],
                            scope=spacecraft.scope)

    def feed(self, stage, propellant, position=None, external=True, **kwargs):
        if external:
            kwargs['nvg'] = self
        if isinstance(position, int):
            quantity = len(self.propellants[stage])
            if position < quantity:
                self.propellants[stage][position] = (propellant, kwargs)
            else:
                for p in range(quantity, position):
                    self.propellants[stage].append((None, 'False'))
        self.propellants[stage].append((propellant, kwargs))

    def _airDrop(self, stage, n, **kwargs):
        # kwargs.update(self.propellants[stage][n][2])
        self.propellants[stage][n] = self.propellants[stage][n][0:2] + (kwargs,)

    def _triggerStage(self, curr_stage):
        self.stage = curr_stage
        for ppl, kwargs in self.propellants[curr_stage]:
            ppl(**kwargs)

    def cruise(self, exit, feedback=None, thruster=None, **entrance):
        '''
        exit: which node to output
        feedback: loss name
        thruster: optimizer name
        entrance: input (formated as keyword arguments)
        '''
        ent = {}
        for k in entrance.keys():
            ent[self.sc.layout[k]] = entrance[k]

        if thruster is None:
            ext = exit
        else:
            ext = exit + (thruster,)
        ext = [self.sc.layout[k] for k in ext]
        tmp = self.sess.run(ext, ent)

        ret = {}
        for i, key in enumerate(exit):
            ret[key] = tmp[i]
        return ret

    def execute(self):
        root_stage = self.stage
        for stg in Stage.groups[root_stage]:
            self._triggerStage(stg)
        self.stage = root_stage

    def _scribe(self):
        if self.flag_empty:
            bp_cfg = {}
        else:
            bp_cfg = self.sc.blueprint.param
        config = {'NG': self.sc.engine.param, 'TM': self.tm.param,
                  'FD': {}, 'DB': self.db.param, 'BP': bp_cfg}
        ft = self.fd.tanks
        for k in ft.keys():
            config['FD'][k] = ft[k].param

        temp_config = os.path.join(os.getcwd(), 'temp_config.json')
        if os.path.exists(temp_config):
            hyper_param = parseConfig(temp_config)
            config['SC'] = hyper_param
            if not os.path.exists(self.tm.param['save_path']):
                os.mkdir(self.tm.param['save_path'])
            recordConfig(os.path.join(self.tm.param['save_path'], 'config.json'), config)
            os.remove(os.path.join(os.getcwd(), 'temp_config.json'))

    def log(self, gauge=True, tachograph=True, blueprint=True, setting=True):
        self.feed(Stage.END, self.db._log, external=False, gauge=gauge, tachograph=tachograph)

        if blueprint:
            self.feed(Stage.START, self.sc.blueprint.log, external=False, log_path=self.db.param['log_path'])
        if setting:
            self.feed(Stage.START, self._scribe, external=False)

    def launch(self):
        try:
            print('+' + 35 * '-' + '+')
            print('| The Spacecraft is being launched. |')
            print('+' + 35 * '-' + '+')
            self._triggerStage(Stage.START)
            self._triggerStage(Stage.RUN)
            self._triggerStage(Stage.END)
            print('+' + 44 * '-' + '+')
            print('| The Spacecraft has arrived at destination. |')
            print('+' + 44 * '-' + '+')

        except BaseException as e:
            if Stage.BREAK_DOWN in self.propellants:
                self._triggerStage(Stage.BREAK_DOWN)
            else:
                raise e